import logging

from typing import List, Dict, Any, Callable, Union, TypedDict, Optional

import functools
import shortuuid

from ppa import ui


logger = logging.getLogger(__name__)


SingleRecord = Dict[str, Any]
ListOfRecords = List[Dict[str, Any]]
AnyRecords = Union[SingleRecord, ListOfRecords]
TableHeader = List[str]
TableRow = List[str]
TableData = List[TableRow]
TabbedTableRaw = Dict[str, AnyRecords]
TableTab = AnyRecords
TabbedTableData = Dict[str, TableData]
EmptyTable = list  # This is not useful as an annotation but it makes return types a bit easier to read.


class ValidationError(Exception):
    pass


class InvalidTableParameters(Exception):
    pass


class NotEnoughRecords(Exception):
    pass


class NoRecordsSupplied(Exception):
    pass


class TableSelection(TypedDict):
    all: ListOfRecords
    first: Optional[SingleRecord]
    total: int


def _validate_flat(records: Any) -> AnyRecords:
    logger.debug(f"validating table records: {records}")
    if not isinstance(records, (list, dict)):
        logger.error(f"invalid table records: {records}")
        raise ValidationError(
            f"Table records must be supplied as a single or list of dictionaries, received a {type(records)} instead."
        )
    if isinstance(records, list):
        invalid_inner_records = [r for r in records if not isinstance(r, dict)]
        if invalid_inner_records:
            logger.error(f"invalid table records: {invalid_inner_records}")
            raise ValidationError(
                "Table records must be supplied as dictionaries, received "
                f"{len(invalid_inner_records)} in a different format. "
                "Please view the logs for more information."
            )
    return records


def _validate_tabbed(records: Any) -> TabbedTableRaw:
    logger.debug(f"validating tabbed table records: {records}")
    if not isinstance(records, dict):
        raise ValidationError(f"Table tabs must be supplied as a dictionary, received a {type(records)} instead.")
    for _, tab_records in records.items():
        if not isinstance(tab_records, (list, dict)):
            logger.error(f"invalid tabbed table records: {tab_records}")
            raise ValidationError(
                f"Table tab contents must be a single or list of dictionaries, received a {type(tab_records)} instead. "
                "Please view the logs for more information."
            )
        if isinstance(tab_records, list):
            invalid_inner_records = [r for r in tab_records if not isinstance(r, dict)]
            if invalid_inner_records:
                logger.error(f"invalid table tab rows: {invalid_inner_records}")
                raise ValidationError(
                    "Each table tab row must be supplied as a dictionary, "
                    f"received {len(invalid_inner_records)} in a different format. "
                    "Please view the logs for more information."
                )
    return records


def _to_list(records: AnyRecords) -> ListOfRecords:
    return records if isinstance(records, list) else [records]


def _process_extract(extract: Union[List[str], Callable]) -> Callable:
    if isinstance(extract, list):
        keys: list = extract
        invalid_keys = [k for k in keys if not isinstance(k, str)]
        if invalid_keys:
            raise Exception("The following keys are not strings: '{}'.".format(", ".join(invalid_keys)))
        return functools.partial(_get_cell_values, keys=keys)

    if callable(extract):
        extractor: Callable = extract
        if extract.__code__.co_argcount != 1:
            raise Exception(
                "Row generator functions must only take a single argument. "
                f"The supplied function takes {extractor.__code__.co_argcount}."
            )
        return extractor

    raise Exception(f"Invalid argument of type {type(extract)}. Please supply a list of keys or a function.")


def _get_cell_values(record: SingleRecord, keys: List[str]) -> TableRow:
    # The word 'null' looks bad in table cells, so if a value is None replace it with an empty string.
    return [str(record.get(k)) if record.get(k) is not None else "" for k in keys]


def _generate_flat(records: AnyRecords, extract: Callable) -> Union[TableData, EmptyTable]:
    filtered_records = list(filter(lambda r: len(r) > 0, records))  # Filter out any empty dictionaries.
    if not filtered_records:
        return []
    return [extract(record) for record in filtered_records]


def _generate_tabbed(tabs: TabbedTableRaw, extract: Callable) -> Union[TabbedTableData, EmptyTable]:
    filtered_tabs = {key: value for key, value in tabs.items() if len(value) > 0}  # Filter out any empty tabs.
    if not filtered_tabs:
        return []
    return {title: [extract(item) for item in _to_list(contents)] for title, contents in filtered_tabs.items()}


def output_table(
    text: str,
    *,
    header: TableHeader,
    extract: Union[List[str], Callable],
    records: AnyRecords,
    # A function that takes a dictionary, a list of keys, & returns a list of table cells.
) -> None:
    ui.output_table(
        text, header=header, data=_generate_flat(_to_list(_validate_flat(records)), _process_extract(extract))
    )


def output_tabbed_table(
    text: str,
    *,
    header: TableHeader,
    extract: Union[List[str], Callable],
    tabs: TabbedTableRaw,  # A dictionary with a string key & a single or list of dictionaries as the value.
) -> None:
    ui.output_table(text, header=header, data=_generate_tabbed(_validate_tabbed(tabs), _process_extract(extract)))


def input_table(
    text: str,
    *,
    header: List[str],
    extract: Union[List[str], Callable],
    records: AnyRecords,
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
    # A function that takes a dictionary, a list of keys, & returns a list of table cells.
) -> TableSelection:
    table_id = str(shortuuid.uuid())
    error_id = str(shortuuid.uuid())
    record_count = len(records)

    # Catch the scenario where the minimum is greater than the maximum.
    if all([minimum is not None, maximum is not None]):
        if minimum > maximum:
            ui.output_error(
                "Table Selection Error: The task has asked for a minimum "
                f"of {minimum} & a maximum of {maximum} selections."
            )
            raise InvalidTableParameters(
                f"Input table was called with a minimum of {minimum} & a maximum of {maximum} selections."
            )

    # Minimum is > 0 but there are 0 items to select from.
    if minimum is not None and minimum > 0:
        if record_count == 0:
            if maximum is not None and minimum == maximum:
                error = (
                    f"Table Selection Error: The task has asked for {minimum} "
                    f"{'selection' if minimum == 1 else 'selections'}, "
                    "but there are no items to select from."
                )
                exception_message = (
                    f"Input table was called with a minimum of {minimum} & "
                    f"maximum of {maximum} selections, but no items were supplied."
                )
            else:
                error = (
                    f"Table Selection Error: The task has asked for at least {minimum} selections, "
                    "but there are no items to select from."
                )
                exception_message = (
                    f"Input table was called with a minimum of {minimum} selections, but no items were supplied."
                )
            ui.output_error(error)
            raise NoRecordsSupplied(exception_message)

        # Now catch other cases where more than 0 items were supplied but the minimum is still a greater value.
        if minimum > record_count:
            ui.output_error(
                f"Table Selection Error: The task has asked for at least {minimum} "
                f"{'selections' if minimum > 1 else 'selection'} from only {record_count} "
                f"{'item(s)' if record_count > 1 else 'item'}."
            )
            raise NotEnoughRecords(
                f"Input table was called with a minimum of {minimum} selections "
                f"but only {record_count} items were supplied."
            )

    records = _to_list(_validate_flat(records))
    table_data = _generate_flat(records, _process_extract(extract))
    while True:
        indices = ui.input_table(text, header=header, data=table_data, element_id=table_id)
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s).", element_id=error_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s).", element_id=error_id)
            continue
        break

    selection = [records[i] for i in indices]
    return TableSelection(total=count, first=selection[0] if count > 0 else None, all=selection)
