import datetime
import logging
import math

from typing import Union, Optional, TypedDict

from . import exceptions

logger = logging.getLogger(__name__)


class TimeUnits(TypedDict):
    days: int
    hours: int
    minutes: int
    seconds: int


class TranslatedTimestamp:
    def __init__(self, unix_epoch: int, windows_epoch: int):
        self.unix_epoch = unix_epoch
        self.windows_epoch = windows_epoch
        self.date_object = datetime.datetime.utcfromtimestamp(unix_epoch)

    def __str__(self):
        return str(self.date_object)

    def __repr__(self):
        return f"TranslatedTimestamp(unix_date={self.__str__()}, windows_epoch={self.windows_epoch})"

    def __eq__(self, other):
        if isinstance(other, datetime.datetime):
            return self.date_object == other
        return str.__eq__(str(self), str(other))

    def __ne__(self, other):
        return not self.__eq__(other)


def windows_bytes_to_unix_date(raw_timestamp: bytes) -> Union[TranslatedTimestamp, None]:
    # Creates a TranslatedTimestamp instance from an LDAP windows epoch timestamp. The raw
    # epoch arrives as bytes as it won't have been decoded from the LDAP response yet.
    if not isinstance(raw_timestamp, bytes):
        raise exceptions.AttributeDecodingError(
            f"Can only decode windows timestamp from bytes. Received {type(raw_timestamp)}."
        )

    try:
        windows_epoch = int(raw_timestamp.decode("utf-8"))
    except (ValueError, AttributeError):
        raise exceptions.AttributeDecodingError(f"Failed to decode windows timestamp with value {raw_timestamp}.")

    if windows_epoch in {9_223_372_036_854_775_807, 0}:  # never expires or not set
        return None

    # Convert AD value (100 nano-second intervals since 1/1/1601) to seconds
    seconds = int(windows_epoch / 10_000_000)

    # Subtract the difference in seconds between AD epoch (1/1/1601) and UNIX epoch (1/1/1970) so we can
    # create a datetime object from a UNIX epoch.
    unix_epoch = int(seconds) - 11_644_473_600

    try:
        return TranslatedTimestamp(unix_epoch, windows_epoch)
    except (ValueError, OverflowError) as e:
        if "out of range" in str(e):
            # A 'never expires' or 'never set' value in AD will be a timestamp in the distant future. In python-land
            # these values should be None. This shouldn't happen because we catch 0 and 9223372036854775807 further
            # up, but better safe than sorry.
            logger.debug(f"UNIX epoch stamp {unix_epoch} is too far in the future to convert.")
            return None
        raise e


def seconds_to_units(seconds: int) -> TimeUnits:
    # Takes a number of seconds and calculates the number of days, hours, minutes, & seconds. This is mainly
    # for making a datetime.timedelta more useful, as they don't have many of these as class attributes.

    # UNIT_DIVISORS contains the division numbers for days, hours, minutes, seconds
    UNIT_DIVISORS = [60 * 60 * 24, 60 * 60, 60, None]

    def _calculate(remainder, divisor_index=0):
        divisor = UNIT_DIVISORS[divisor_index]
        if divisor is not None:
            units = math.floor(remainder / divisor)
            yield units
        else:
            yield int(remainder)
            return
        yield from _calculate(remainder - (units * divisor), divisor_index + 1)

    return TimeUnits(**dict(zip(["days", "hours", "minutes", "seconds"], _calculate(seconds))))  # type: ignore


def get_windows_time_delta(raw_timestamp: bytes) -> Optional[TimeUnits]:
    # Used to generate & return a timedelta representing the difference between windows EPOCH & a supplied windows
    # timestamp. This was introduced to support the maxPwdAge root domain attribute, whose value is the number of
    # 100 nanosecond intervals since AD epoch (1/1/1601). Sometimes these are negative ints which need converting
    # to positive ints before they're processed.
    #
    # For example...
    #
    # If the value coming into this function is b'-43200000000000':
    #
    # - It will be decoded to utf-8, then turned into a a positive int
    # - The int is then divided by 10_000_000 (converting nano-second intervals to second intervals)
    # - That is now the number of seconds between windows epoch & this supplied timestamp
    #
    # A datetime object representing the supplied timestamp is created by adding the calculated number
    # of seconds to the windows epoch datetime object. Lastly a timedelta object is created by comparing
    # both the windows epoch & supplied timestamp datetime objects.
    #
    # Using the example value of -43200000000000, the outputted timedelta will be:
    #
    # datetime.timedelta(days=50)
    #
    # This matches the value displayed in Active Directory & can now be used to calculate information
    # such as how long a user has before their password expires.
    if not isinstance(raw_timestamp, bytes):
        raise exceptions.AttributeDecodingError(
            f"Can only decode windows timestamp from bytes. Received {type(raw_timestamp)}."
        )

    # Create a datetime representing windows epoch so we can compare the supplied timestamp against it.
    windows_epoch = datetime.datetime(1601, 1, 1)

    logger.debug(f"converting active directory timestamp {raw_timestamp} to an int")
    try:
        timestamp = int(raw_timestamp.decode("utf-8"))
    except (ValueError, AttributeError):
        raise exceptions.AttributeDecodingError(f"Failed to decode windows timestamp with value {raw_timestamp}.")

    if timestamp < 0:
        # The maxPwdAge attribute is, for some reason, a negative long integer.
        timestamp = timestamp * -1

    if timestamp in {9_223_372_036_854_775_807, 9_223_372_036_854_775_808, 0}:  # never expires or not set
        return None

    seconds_difference = int(timestamp / 10_000_000)
    timestamp_date_object = windows_epoch + datetime.timedelta(seconds=seconds_difference)
    final_delta = timestamp_date_object - windows_epoch
    logger.debug(f"final delta: {int(final_delta.total_seconds())}")
    return seconds_to_units(int(final_delta.total_seconds()))
