from typing import Optional, List, Dict

from . import users, groups, computers, _helpers
from ._object_classes import User, Group, Computer


def get_user(
    connection,
    *,
    search_attribute: Optional[str] = None,
    search_base: Optional[str] = None,
    exclude: Optional[Dict[str, str]] = None,
    auto_select: Optional[bool] = False,
    auto_wildcard: Optional[bool] = False,
    title: Optional[str] = None,
) -> User:
    """An interactive search for an Active Directory user. If multiple records are
    found, the PPA user will be asked to select a single user from a table.

    User records can be excluded from the search by using the `exclude` keyword argument.

    :param connection: An LDAPS connection.
    :param search_attribute: The LDAP attribute to search against. If not supplied, the PPA user
        will be presented with a dropdown of user attributes.
    :param search_base: An optional distinguishedName to use as the root of the search. Defaults
        to the root of the domain.
    :param exclude: A dictionary of exclusion criteria. Each key should be an LDAP attribute, and
        each value should be a regular expression. A `match <https://docs.python.org/3.8/library/re.html#re.match>`_
        will be attempted using the supplied criteria against every record. Records that produce
        at least one match will be excluded. See **code example 3** for usage information.
    :param auto_select: Optional bool. Defaults to `False`. If a search returns a single record, the user will be asked
        to confirm before returning the single record. To auto-select from a single record search result, set
        `auto_select` to `True`.
    :param auto_wildcard: Optional bool. Defaults to `False`. Set to `True` to automatically add a `*` wildcard to the
        end of each search value. This can be useful when users are not familiar with having to use wildcards and expect
        a `search` to behave this way by default.
    :param title: Optional str. Defaults to `Get Active Directory User`. The header displayed to the Task Operator when
        the search helper starts.

    :raises exceptions.NoSuchObject: if no users are found within 10 attempts.
    :raises exceptions.ParameterError: if a key in `exclude` is either invalid or missing on an object.
    :raises exceptions.ParameterError: if a value in `exclude` is an invalid regular expression.

    :returns: A single Active Directory user object.

    **Example 1** - Searching with no supplied arguments. The PPA user will be prompted to choose
    both a search attribute and search value:

            .. code-block:: python

                helpers.get_user(connection)

    **Example 2** - Searching against supplied attribute `userPrincipalName` with a user prompted value,
    excluding `Administrator` and any users from the `Directors` Organizational Unit:

            .. code-block:: python

                helpers.get_user(
                    connection,
                    search_attribute="userPrincipalName",
                    exclude={
                        "distinguishedName": ".*OU=Directors,.*"
                        "sAMAccountName": "Administrator"
                    }
                )


    .. note:: If a search value provided by the PPA user contains a wildcard character, that is not at the
        end of the string, it will be escaped.
    .. seealso:: The documentation for `re.match <https://docs.python.org/3.8/library/re.html#re.match>`_ and
        `match vs search <https://docs.python.org/3/library/re.html#search-vs-match>`_
        for more information on filtering.
    """
    return _helpers._get_object(
        connection,
        "User Account",
        [
            {"display": "SAM Account Name", "value": "sAMAccountName"},
            {"display": "Email Address", "value": "mail"},
            {"display": "User Principal Name", "value": "userPrincipalName"},
        ],
        users.search,
        search_attribute=search_attribute,
        search_base=search_base,
        exclude=exclude,
        auto_select=auto_select,
        auto_wildcard=auto_wildcard,
        title=title,
    )


def get_group(
    connection,
    *,
    search_attribute: Optional[str] = None,
    search_base: Optional[str] = None,
    exclude: Optional[Dict[str, str]] = None,
    auto_select: Optional[bool] = False,
    auto_wildcard: Optional[bool] = False,
    title: Optional[str] = None,
) -> Group:
    """An interactive search for an Active Directory group. If multiple records are
    found, the PPA user will be asked to select a single group from a table.

    Group records can be excluded from the search by using the `exclude` keyword argument.

    :param connection: An LDAPS connection.
    :param search_attribute: The LDAP attribute to search against. If not supplied, the PPA user
        will be presented with a dropdown of group attributes.
    :param search_base: An optional distinguishedName to use as the root of the search. Defaults
        to the root of the domain.
    :param exclude: A dictionary of exclusion criteria. Each key should be an LDAP attribute, and
        each value should be a regular expression. A `match <https://docs.python.org/3.8/library/re.html#re.match>`_
        will be attempted using the supplied criteria against every record. Records that produce
        at least one match will be excluded. See **code example 3** for usage information.
    :param auto_select: Optional bool. Defaults to `False`. If a search returns a single record, the user will be asked
        to confirm before returning the single record. To auto-select from a single record search result, set
        `auto_select` to `True`.
    :param auto_wildcard: Optional bool. Defaults to `False`. Set to `True` to automatically add a `*` wildcard to the
        end of each search value. This can be useful when users are not familiar with having to use wildcards and expect
        a `search` to behave this way by default.
    :param title: Optional str. Defaults to `Get Active Directory Group`. The header displayed to the Task Operator when
        the search helper starts.

    :raises exceptions.NoSuchObject: if no group are found within 10 attempts.
    :raises exceptions.ParameterError: if a key in `exclude` is either invalid or missing on an object.
    :raises exceptions.ParameterError: if a value in `exclude` is an invalid regular expression.

    :returns: A single Active Directory group object.

    **Example 1** - Searching with no supplied arguments. The PPA user will be prompted to choose
    both a search attribute and a search value:

            .. code-block:: python

                helpers.get_group(connection)

    **Example 2** - Searching against supplied attribute `cn`. This will only prompt the PPA user for
    a search value:

            .. code-block:: python

                helpers.get_group(connection, search_attribute="cn")

    **Example 3** - Searching against supplied attribute `cn`, prompting the PPA user for a value,
    excluding `Domain Admins` and any groups in the `Directors` Organizational Unit from the search:

            .. code-block:: python

                helpers.get_group(
                    connection,
                    search_attribute="cn",
                    exclude={
                        "distinguishedName": ".*OU=Directors,.*"
                        "cn": "Domain Admins"
                    }
                )


    .. note:: If a search value provided by the PPA user contains a wildcard character, that is not at the
        end of the string, it will be escaped.
    .. seealso:: The documentation for `re.match <https://docs.python.org/3.8/library/re.html#re.match>`_ and
        `match vs search <https://docs.python.org/3/library/re.html#search-vs-match>`_
        for more information on filtering.
    """
    return _helpers._get_object(
        connection,
        "Security Group",
        [{"display": "SAM Account Name", "value": "sAMAccountName"}, {"display": "Common Name", "value": "cn"}],
        groups.search,
        search_attribute=search_attribute,
        search_base=search_base,
        exclude=exclude,
        auto_select=auto_select,
        auto_wildcard=auto_wildcard,
        title=title,
    )


def get_computer(
    connection,
    *,
    search_attribute: Optional[str] = None,
    search_base: Optional[str] = None,
    exclude: Optional[Dict[str, str]] = None,
    auto_select: Optional[bool] = False,
    auto_wildcard: Optional[bool] = False,
    title: Optional[str] = None,
) -> Computer:
    """An interactive search for an Active Directory computer. If multiple records are
    found, the PPA user will be asked to select a single user from a table.

    Computer records can be excluded from the search by using the `exclude` keyword argument.

    :param connection: An LDAPS connection.
    :param search_attribute: The LDAP attribute to search against. If not supplied, the PPA user
        will be presented with a dropdown of user attributes.
    :param search_base: An optional distinguishedName to use as the root of the search. Defaults
        to the root of the domain.
    :param exclude: A dictionary of exclusion criteria. Each key should be an LDAP attribute, and
        each value should be a regular expression. A `match <https://docs.python.org/3.8/library/re.html#re.match>`_
        will be attempted using the supplied criteria against every record. Records that produce
        at least one match will be excluded. See **code example 3** for usage information.
    :param auto_select: Optional bool. Defaults to False. If a search returns a single record, the user will be asked
        to confirm before returning the single record. To auto-select from a single record search result, set
        `auto_select` to `True`.
    :param auto_wildcard: Optional bool. Defaults to `False`. Set to `True` to automatically add a `*` wildcard to the
        end of each search value. This can be useful when users are not familiar with having to use wildcards and expect
        a `search` to behave this way by default.
    :param title: Optional str. Defaults to `Get Active Directory Computer`. The header displayed to the Task Operator
        when the search helper starts.

    :raises exceptions.NoSuchObject: if no computers are found within 10 attempts.
    :raises exceptions.ParameterError: if a key in `exclude` is either invalid or missing on an object.
    :raises exceptions.ParameterError: if a value in `exclude` is an invalid regular expression.

    :returns: A single Active Directory computer object.

    **Example 1** - Searching with no supplied arguments. The PPA user will be prompted to choose
    both a search attribute and search value:

            .. code-block:: python

                helpers.get_computer(connection)

    **Example 2** - Searching against supplied attribute `userPrincipalName` with a user prompted value,
    excluding any computers from the `Domain Controllers` Organizational Unit:

            .. code-block:: python

                helpers.get_computer(
                    connection,
                    search_attribute="userPrincipalName",
                    exclude={
                        "distinguishedName": ".*OU=Domain Controllers,.*"
                    }
                )


    .. note:: If a search value provided by the PPA user contains a wildcard character, that is not at the
        end of the string, it will be escaped.
    .. seealso:: The documentation for `re.match <https://docs.python.org/3.8/library/re.html#re.match>`_ and
        `match vs search <https://docs.python.org/3/library/re.html#search-vs-match>`_
        for more information on filtering.
    """
    return _helpers._get_object(
        connection,
        "Computer",
        [
            {"display": "SAM Account Name", "value": "sAMAccountName"},
            {"display": "Display Name", "value": "displayName"},
            {"display": "Common Name", "value": "cn"},
        ],
        computers.search,
        search_attribute=search_attribute,
        search_base=search_base,
        exclude=exclude,
        auto_select=auto_select,
        auto_wildcard=auto_wildcard,
        title=title,
    )


def add_user_to_groups(
    connection,
    *,
    distinguished_name: str,
    search_attribute: Optional[str] = None,
    exclude: Optional[Dict[str, str]] = None,
    auto_select: Optional[bool] = False,
) -> tuple:
    """ Add a user to one or more security groups.

    This function allows the PPA user to search for and select multiple security groups. The supplied
    user will then be added to all groups, after the selection has been confirmed.

    Group records can be excluded from the search by using the `exclude` keyword argument.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.
    :param search_attribute: The LDAP attribute to search against. If not supplied, the PPA user will be
        presented with a dropdown of group attributes.
    :param exclude: A dictionary of exclusion criteria. Each key should be an LDAP attribute, and
        each value should be a regular expression. A `match <https://docs.python.org/3.8/library/re.html#re.match>`_
        will be attempted using the supplied criteria against every record. Records that produce
        at least one match will be excluded. See the **code example** for usage information.
    :param auto_select: Optional bool. Defaults to False. If a search returns a single record, the user will be asked
        to confirm before returning the single record. To auto-select from a single record search result, set
        `auto_select` to `True`.

    :raises exceptions.NoSuchObject: if the supplied user does not exist.
    :raises exceptions.ParameterError: if a key in `exclude` is either invalid or missing on an object.
    :raises exceptions.ParameterError: if a value in `exclude` is an invalid regular expression.

    :returns: A 2 element tuple containing lists of the groups the user was added to, and any groups the
        user was already a member of. Format: `([groups_added_to], [existing_memberships])`

    **Example** - Adding the supplied user to groups. The PPA user will be prompted to choose a search
    attribute, after which they can search for and choose multiple groups. All groups from the `Directors`
    Organizational Unit will be excluded:

        .. code-block:: python

            helpers.add_user_to_groups(
                connection,
                distinguished_name="cn=test user,cn=users,dc=test,dc=domain,dc=com",
                exclude={
                    "distinguishedName": ".*OU=Directors,.*"
                }
            )

    .. seealso:: :ref:`active_directory:Helper Functions`, the documentation for `re.match
        <https://docs.python.org/3.8/library/re.html#re.match>`_ and
        `match vs search <https://docs.python.org/3/library/re.html#search-vs-match>`_
        for more information on filtering.
"""
    return _helpers._add_object_to_groups(
        connection,
        users,
        "user",
        distinguished_name=distinguished_name,
        search_attribute=search_attribute,
        exclude=exclude,
        auto_select=auto_select,
    )


def add_computer_to_groups(
    connection,
    *,
    distinguished_name: str,
    search_attribute: Optional[str] = None,
    exclude: Optional[Dict[str, str]] = None,
    auto_select: Optional[bool] = False,
) -> tuple:
    """ Add a computer to one or more security groups.

    This function allows the PPA user to search for and select multiple security groups. The supplied
    computer will then be added to all groups, after the selection has been confirmed.

    Group records can be excluded from the search by using the `exclude` keyword argument.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target computer.
    :param search_attribute: The LDAP attribute to search against. If not supplied, the PPA user will be
        presented with a dropdown of group attributes.
    :param exclude: A dictionary of exclusion criteria. Each key should be an LDAP attribute, and
        each value should be a regular expression. A `match <https://docs.python.org/3.8/library/re.html#re.match>`_
        will be attempted using the supplied criteria against every record. Records that produce
        at least one match will be excluded. See the **code example** for usage information.
    :param auto_select: Optional bool. Defaults to False. If a search returns a single record, the user will be asked
        to confirm before returning the single record. To auto-select from a single record search result, set
        `auto_select` to `True`.

    :raises exceptions.NoSuchObject: if the supplied computer does not exist.
    :raises exceptions.ParameterError: if a key in `exclude` is either invalid or missing on an object.
    :raises exceptions.ParameterError: if a value in `exclude` is an invalid regular expression.

    :returns: A 2 element tuple containing lists of the groups the computer was added to, and any groups the
        computer was already a member of. Format: `([groups_added_to], [existing_memberships])`

    **Example** - Adding the supplied computer to groups. The PPA user will be prompted to choose a search
    attribute, after which they can search for and choose multiple groups. All groups from the `Domain Controllers`
    Organizational Unit will be excluded:

        .. code-block:: python

            helpers.add_computer_to_groups(
                connection,
                distinguished_name="cn=test computer,cn=computers,dc=test,dc=domain,dc=com",
                exclude={
                    "distinguishedName": ".*OU=Domain Controllers,.*"
                }
            )

    .. seealso:: :ref:`active_directory:Helper Functions`, the documentation for `re.match
        <https://docs.python.org/3.8/library/re.html#re.match>`_ and
        `match vs search <https://docs.python.org/3/library/re.html#search-vs-match>`_
        for more information on filtering.
"""
    return _helpers._add_object_to_groups(
        connection,
        computers,
        "computer",
        distinguished_name=distinguished_name,
        search_attribute=search_attribute,
        exclude=exclude,
        auto_select=auto_select,
    )


def remove_user_from_groups(
    connection,
    *,
    distinguished_name: str,
    exclude: Optional[Dict[str, str]] = None,
    auto_select: Optional[bool] = False,
) -> List[Group]:
    """ Remove a user from one or more security groups.

    This function will present a table of the supplied users' group memberships, and
    expect at least one to be chosen. Membership of the 'Domain Users' group
    is ignored.

    The supplied user will then be removed from all chosen groups, after the selection
    has been confirmed.

    Group records can be excluded from the search by using the `exclude` keyword argument.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target user.
    :param exclude: A dictionary of exclusion criteria. Each key should be an LDAP attribute, and
        each value should be a regular expression. A `match <https://docs.python.org/3.8/library/re.html#re.match>`_
        will be attempted using the supplied criteria against every record. Records that produce
        at least one match will be excluded. See the **code example** for usage information.
    :param auto_select: Optional bool. Defaults to False. If a search returns a single record, the user will be asked
        to confirm before returning the single record. To auto-select from a single record search result, set
        `auto_select` to `True`.

    :raises exceptions.NoSuchObject: if the supplied user does not exist.
    :raises exceptions.OperationFailed: if the supplied user is not in any groups.
    :raises exceptions.ParameterError: if a key in `exclude` is either invalid or missing on an object.
    :raises exceptions.ParameterError: if a value in `exclude` is an invalid regular expression.

    :returns: The groups the user was removed from as a list of Active Directory group objects.

    **Example 1** - Removing `test user` from groups, excluding any groups in the `Directors` Organization Unit:

        .. code-block:: python

            helpers.remove_user_from_groups(
                connection,
                distinguished_name="cn=test user,cn=users,dc=test,dc=domain,dc=com",
                exclude={
                    "distinguishedName": ".*OU=Directors,.*"
                }
            )

    .. seealso:: :ref:`active_directory:Helper Functions`, the documentation for `re.match
        <https://docs.python.org/3.8/library/re.html#re.match>`_ and
        `match vs search <https://docs.python.org/3/library/re.html#search-vs-match>`_
        for more information on filtering.
"""
    return _helpers._remove_object_from_groups(
        connection, users, "user", distinguished_name=distinguished_name, exclude=exclude, auto_select=auto_select
    )


def remove_computer_from_groups(
    connection,
    *,
    distinguished_name: str,
    exclude: Optional[Dict[str, str]] = None,
    auto_select: Optional[bool] = False,
) -> List[Group]:
    """ Remove a user from one or more security groups.

    This function will present a table of the supplied computer's group memberships, and
    expect at least one to be chosen. Membership of the 'Domain Users' group
    is ignored.

    The supplied user will then be removed from all chosen groups, after the selection
    has been confirmed.

    Group records can be excluded from the search by using the `exclude` keyword argument.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target computer.
    :param exclude: A dictionary of exclusion criteria. Each key should be an LDAP attribute, and
        each value should be a regular expression. A `match <https://docs.python.org/3.8/library/re.html#re.match>`_
        will be attempted using the supplied criteria against every record. Records that produce
        at least one match will be excluded. See the **code example** for usage information.
    :param auto_select: Optional bool. Defaults to False. If a search returns a single record, the user will be asked
        to confirm before returning the single record. To auto-select from a single record search result, set
        `auto_select` to `True`.

    :raises exceptions.NoSuchObject: if the supplied computer does not exist.
    :raises exceptions.OperationFailed: if the supplied computer is not in any groups.
    :raises exceptions.ParameterError: if a key in `exclude` is either invalid or missing on an object.
    :raises exceptions.ParameterError: if a value in `exclude` is an invalid regular expression.

    :returns: The groups the computer was removed from as a list of Active Directory group objects.

    **Example 1** - Removing `test computer` from groups, excluding any groups in the `Directors` Organization Unit:

        .. code-block:: python

            helpers.remove_computer_from_groups(
                connection,
                distinguished_name="cn=test computer,cn=computers,dc=test,dc=domain,dc=com",
                exclude={
                    "distinguishedName": ".*OU=Directors,.*"
                }
            )

    .. seealso:: :ref:`active_directory:Helper Functions`, the documentation for `re.match
        <https://docs.python.org/3.8/library/re.html#re.match>`_ and
        `match vs search <https://docs.python.org/3/library/re.html#search-vs-match>`_
        for more information on filtering.
"""
    return _helpers._remove_object_from_groups(
        connection,
        computers,
        "computer",
        distinguished_name=distinguished_name,
        exclude=exclude,
        auto_select=auto_select,
    )
