"""
Containers

This module contains actions related to Active Directory Containers.
"""
import logging

from typing import List, Dict, Optional, Union, Any


from ppa import ui
from active_directory import containers, exceptions as sdk_exceptions
from playbook_tools import tables, process

from .exceptions import NoContainerFound, InvalidAttribute
from ._wrappers import ldap_function, idempotent_create, idempotent_remove, process_container, bool_to_search_scope
from .records import DomainController, Container


CN_TABLE_HEADER = ["Name", "Distinguished Name"]
CN_TABLE_FIELDS = ["cn", "distinguishedName"]


logger = logging.getLogger(__name__)


@ldap_function
def get_all(
    domain_controller: DomainController,
    search_base: Optional[str] = None,
    one_level: Optional[bool] = False,
    ignore_system: Optional[bool] = False,
) -> List[Container]:
    """Get all containers.

!!! question "Minimum Plugin Version: 7.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

    - __one_level:__ set to `true` to only search one level (see [here](../advanced_queries/#search-scopes) for more information)

    - __ignore_system:__ set to `true` to ignore the default _System_ container & its contents (defaults to `false`)

??? success "Output"

    A list of [Containers](../inputs_outputs/#container).

??? example
    ```YAML linenums="1"
    - active_directory.containers.get_all:
      load:
        domain_controller: domain_controller
      save: all_containers
    ```
"""
    message = "getting all containers"
    if search_base is not None:
        message = f"{message} starting from {search_base}"
    if one_level is True:
        message = f"{message} at one level"
    logger.info(message)
    return list(
        map(
            process_container,
            containers.get_all(
                domain_controller,
                search_base=search_base,
                search_scope=bool_to_search_scope(one_level),
                ignore_system=ignore_system,
            ),
        )
    )


@ldap_function
def by_distinguishedname(domain_controller: DomainController, distinguishedName: str) -> Container:
    """Get a container by its 'distinguishedName' attribute.

!!! question "Minimum Plugin Version: 7.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ The container's `distinguishedName`

??? success "Output"

    A single [Container](../inputs_outputs/#container).

??? example
    ```YAML linenums="1"
    - active_directory.containers.by_distinguishedname:
        distinguishedName: CN=Engineers,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller
      save: example_container
    ```
"""
    logger.info(f"getting container with distinguishedName '{distinguishedName}'")
    try:
        return process_container(containers.search(domain_controller, distinguishedName=distinguishedName)[0])
    except IndexError:
        raise NoContainerFound(f"No container found with distinguishedName '{distinguishedName}'.")


@ldap_function
def search(
    domain_controller: DomainController,
    search_params: Dict[str, Any],
    search_base: Optional[str] = None,
    one_level: Optional[bool] = False,
) -> process.SearchResult:
    """Search for containers using LDAP attributes & values.

!!! question "Minimum Plugin Version: 7.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller)

    - __search_params:__ a dictionary containing [Container](../inputs_outputs#container) keys & values to use in the search

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

    - __one_level:__ set to `true` to only search one level (see [here](../advanced_queries/#search-scopes) for more information)

??? success "Output"

    A [SearchResult](../../../searches/#search-result-dictionary) containing:

    - **total**: the number of [Containers](../inputs_outputs/#container) found

    - **all**: a list of [Containers](../inputs_outputs/#container) found

    - **first**: the first [Container](../inputs_outputs/#container) found

    - **last**: the last [Container](../inputs_outputs/#container) found

??? warning "Search Result Format"

    The search result format was updated in version 6.0.0 of this plugin.

    It is not compatible with playbooks written against previous versions.

    See [here](/searches/#using-search-results) for more information.

??? example

    Searching for all containers in the __Users__ CN whose `cn` starts with __admin*__:

    ```YAML linenums="1"
    - active_directory.containers.search:
        search_params:
          cn: admin*
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller
      save: containers
    ```

!!! tip

    You can use `*` as a wildcard at the end of search values.
"""
    logger.info(
        f"searching for containers in search base {search_base if search_base else 'default'} using the following parameters: {search_params}"
    )
    return process.search_result(
        list(
            map(
                process_container,
                containers.search(
                    domain_controller, search_base=search_base, search_scope=bool_to_search_scope(one_level), **search_params
                ),
            )
        )
    )


@ldap_function
@idempotent_create
def create(
    domain_controller: DomainController, name: str, distinguishedName: str, extra_params: Optional[Dict[str, Any]] = None,
) -> None:
    """Create a new container.

!!! question "Minimum Plugin Version: 7.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller)

    - __name:__ the new container name

    - __distinguishedName:__ the new container distinguishedName

    - __extra_params:__ a dictionary containing any extra LDAP attributes & values for the container

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.containers.create:
        name: Example CN
        distinguishedName: CN=Example CN,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller
    ```
    """
    logger.info(f"creating container '{name}' at '{distinguishedName}' with the following extra parameters: {extra_params}")
    try:
        return containers.create(domain_controller, distinguished_name=distinguishedName, name=name, **extra_params or {},)
    except sdk_exceptions.ParameterError as e:
        logger.error(e)
        message = "One or more of the following is not a supported container attribute: {}.".format(
            ", ".join([f"'{key}'" for key in list(extra_params.keys())])
        )
        ui.output_error(message)
        raise InvalidAttribute(message)
    except sdk_exceptions.BadDistinguishedName as e:
        logger.exception(e)
        message = f"Failed to create the container as the distinguished name is invalid: '{distinguishedName}'"
        ui.output_error(message)
        raise InvalidAttribute(message)
    except sdk_exceptions.InvalidData as e:
        logger.exception(e)
        message = "Failed to create the container as one or more of the attributes is too long."
        ui.output_error(message)
        raise InvalidAttribute(message)


@ldap_function
@idempotent_remove
def delete(domain_controller: DomainController, distinguishedName: str) -> None:
    """Delete a container.

!!! question "Minimum Plugin Version: 7.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller)

    - __distinguishedName:__ the distinguishedName of the container to delete

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.containers.delete:
        distinguishedName: CN=Example CN,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller
    ```
    """
    logger.info(f"deleting container with distinguished name {distinguishedName}")
    return containers.delete(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def move(domain_controller: DomainController, distinguishedName: str, parent_distinguishedName: str) -> None:
    """Move a container to a different OU or CN.

!!! question "Minimum Plugin Version: 7.3.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller)

    - __distinguishedName:__ the distinguishedName of the container to move

    - __parent_distinguishedName:__ the distinguishedName of the new container parent

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - active_directory.containers.move:
        distinguishedName: CN=Example CN,DC=Example,DC=Domain
        parent_distinguishedName: OU=Staff,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller
    ```
    """
    logger.info(f"moving container with distinguished name {distinguishedName} to location {parent_distinguishedName}")
    return containers.move(domain_controller, distinguished_name=distinguishedName, new_parent=parent_distinguishedName)


def display(text: str, containers: Union[Container, List[Container]]) -> None:
    """Display containers in a table.

The table will have the following columns:

- Name
- Distinguished Name

!!! question "Minimum Plugin Version: 7.3.0"

??? info "Input"

    - __text:__ the title of the table

    - __containers:__ any number of [Containers](../inputs_outputs/#container)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - active_directory.containers.get_all:
      load:
        domain_controller: domain_controller
      save: all_containers

    - active_directory.containers.display:
        text: All Containers
      load:
        containers: all_containers
    ```
    """
    tables.output_table(text, header=CN_TABLE_HEADER, extract=CN_TABLE_FIELDS, records=containers)


def display_tabbed(text: str, tabs: Dict[str, Union[Container, List[Container]]]) -> None:
    """Display containers in a table with multiple tabs.

The table will have the following columns:

- Name
- Distinguished Name

!!! question "Minimum Plugin Version: 7.3.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is any number of [Containers](../inputs_outputs/#container)

??? success "Output"

    Nothing is outputted by this action.

??? example

    Auditing user accounts from 2 containers & presenting them in a tabbed table:

    ```YAML linenums="1"
    - set:
        name: tabs
        value: {}

    - active_directory.containers.get_all:
      load:
        domain_controller: production_domain
      save: production_containers

    - active_directory.containers.get_all:
      load:
        domain_controller: uat_domain
      save: uat_containers

    - ppa_tools.dictionaries.insert:
        name: Production
      load:
        value: production_containers
        dictionary: tabs
      save: tabs

    - ppa_tools.dictionaries.insert:
        name: UAT
      load:
        value: uat_containers
        dictionary: tabs
      save: tabs

    - active_directory.containers.display_tabbed:
        text: Containers
      load:
        tabs: tabs
    ```
    """
    tables.output_tabbed_table(text, header=CN_TABLE_HEADER, extract=CN_TABLE_FIELDS, tabs=tabs)


def select(
    text: str, containers: Union[Container, List[Container]], minimum: Optional[int] = None, maximum: Optional[int] = None,
) -> tables.TableSelection:
    """Display containers in a table & prompt the task operator to make a selection.

The table will have the following columns:

- Name
- Distinguished Name

!!! question "Minimum Plugin Version: 7.3.0"

??? info "Input"

    - __text:__ the title of the table

    - __containers:__ any number of [Containers](../inputs_outputs/#container)

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Containers](../inputs_outputs/#container)

    - **all**: a list of selected [Containers](../inputs_outputs/#container)

    - **first**: the first selected [Container](../inputs_outputs/#container)

??? example

    ```YAML linenums="1"
    - active_directory.containers.get_all:
      load:
        domain_controller: domain_controller
      save: containers

    - active_directory.containers.select:
        text: Select Containers
      load:
        containers: containers
      save: selection
    ```
    """
    return tables.input_table(
        text, header=CN_TABLE_HEADER, extract=CN_TABLE_FIELDS, records=containers, minimum=minimum, maximum=maximum
    )


def select_one(text: str, containers: Union[Container, List[Container]]) -> Container:
    """Display Containers in a table & prompt the task operator to select one.

The table will have the following columns:

- Name
- Distinguished Name

!!! question "Minimum Plugin Version: 7.3.0"

??? info "Input"

    - __text:__ the title of the table

    - __containers:__ any number of [Containers](../inputs_outputs/#container)

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A single [Container](../inputs_outputs/#container).

??? example

    ```YAML linenums="1"
    - active_directory.containers.get_all:
      load:
        domain_controller: domain_controller
      save: all_containers

    - active_directory.containers.select_one:
        text: Select an Organizational Unit
      load:
        containers: all_containers
      save: container
    ```
    """
    return select(text, containers, minimum=1, maximum=1)["first"]


@ldap_function
def exists(domain_controller: DomainController, **search_params: Union[str, int, None]) -> bool:
    """Search for containers using LDAP attributes & values to identify if any were found.

!!! question "Minimum Plugin Version: 7.3.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs#domaincontroller) dictionary

    - __search_params:__ a dictionary containing [Container keys](../inputs_outputs#container) & values to use in the search

??? success "Output"

    A boolean is outputted by this action

    - `true` if one or more containers are found

    - `false` if no containers are found


??? example

    Searching using a unique attribute:

    ```YAML linenums="1"
    - active_directory.containers.exists:
        distinguishedName: CN=Engineers,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller
      save: item_exists
    ```

!!! tip "Wildcard Searching"

    It is possible to use `*` as a wildcard at the __end__ of search values, but this can make the search slow.

    """
    return bool(containers.search(domain_controller, **search_params))
